package org.xbib.event;

import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.SubscriberExceptionContext;
import org.xbib.event.bus.SubscriberExceptionHandler;
import org.xbib.event.clock.ClockEventManager;
import org.xbib.event.io.file.FileFollowEventManager;
import org.xbib.event.io.path.PathEventManager;
import org.xbib.event.syslog.SyslogEventManager;
import org.xbib.event.timer.TimerEventManager;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.time.ZoneId;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public final class EventManager implements Closeable {

    private static final Logger logger = Logger.getLogger(EventManager.class.getName());

    private static ExecutorService executorService;

    private static ClassLoader classLoader;

    private static AsyncEventBus eventBus;

    private static List<EventConsumer> eventConsumers;

    static {
        createInstances();
    }

    private final ClockEventManager clockEventManager;

    private final TimerEventManager timerEventManager;

    private final FileFollowEventManager fileFollowEventManager;

    private final PathEventManager pathEventManager;

    private final SyslogEventManager syslogEventManager;

    private EventManager(Settings settings) {
        this.clockEventManager = new ClockEventManager(settings, eventBus, classLoader);
        this.timerEventManager = new TimerEventManager(settings, eventBus, classLoader, ZoneId.systemDefault());
        this.fileFollowEventManager = new FileFollowEventManager(settings, eventBus, executorService, classLoader);
        this.pathEventManager = new PathEventManager(settings, eventBus, executorService, classLoader);
        this.syslogEventManager = new SyslogEventManager(settings, eventBus);
    }

    public static EventManager newEventManager(Settings settings) {
        return new EventManager(settings);
    }

    public static void register(EventConsumer eventConsumer) {
        Objects.requireNonNull(eventConsumer, "event consumer must not be null");
        eventConsumers.add(eventConsumer);
        eventBus.register(eventConsumer);
    }

    public List<EventConsumer> getEventConsumers() {
        return eventConsumers;
    }

    public ClockEventManager getClockEventManager() {
        return clockEventManager;
    }

    public TimerEventManager getTimerEventManager() {
        return timerEventManager;
    }

    public FileFollowEventManager getFileFollowEventManager() {
        return fileFollowEventManager;
    }

    public PathEventManager getPathEventManager() {
        return pathEventManager;
    }

    public SyslogEventManager getSyslogEventManager() {
        return syslogEventManager;
    }

    @Override
    public boolean equals(Object obj) {
        return super.equals(obj);
    }

    @Override
    public void close() throws IOException {
        for (EventConsumer eventConsumer : eventConsumers) {
            eventConsumer.close();
        }
        clockEventManager.close();
        timerEventManager.close();
        fileFollowEventManager.close();
        pathEventManager.close();
        syslogEventManager.close();
    }

    private static class EventManagerExceptionHandler implements SubscriberExceptionHandler {
        @Override
        public void handleException(Throwable exception, SubscriberExceptionContext context) {
            logger.log(Level.SEVERE, exception.getMessage(), exception);
        }
    }

    private static void createInstances() {
        if (executorService == null) {
            executorService = Executors.newFixedThreadPool(2);
        }
        if (classLoader == null) {
            classLoader = EventManager.class.getClassLoader();
        }
        if (eventBus == null) {
            eventBus = new AsyncEventBus(executorService, new EventManagerExceptionHandler());
        }
        if (eventConsumers == null) {
            eventConsumers = new ArrayList<>();
        }
    }
}
