package org.xbib.event.clock;

import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.EventBus;
import org.xbib.settings.Settings;
import org.xbib.time.schedule.CronExpression;
import org.xbib.time.schedule.CronSchedule;

import java.io.Closeable;
import java.io.IOException;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ThreadFactory;
import java.util.logging.Level;
import java.util.logging.Logger;

public class ClockEventManager implements Closeable {

    private static final Logger logger = Logger.getLogger(ClockEventManager.class.getName());

    private final CronSchedule<Integer> cronSchedule;

    public ClockEventManager(Settings settings) {
        this(settings, new AsyncEventBus(Executors.newSingleThreadExecutor()), ClockEventManager.class.getClassLoader());
    }

    public ClockEventManager(Settings settings,
                             EventBus eventBus,
                             ClassLoader classLoader) {
        ThreadFactory threadFactory = new ThreadFactory() {
            int n = 1;
            @Override
            public Thread newThread(Runnable r) {
                return new Thread(r, "clock-event-manager-" + (n++));
            }
        };
        ScheduledExecutorService executorService =
                Executors.newScheduledThreadPool(settings.getAsInt("pool.size", 2), threadFactory);
        this.cronSchedule = new CronSchedule<>(executorService);
        for (Map.Entry<String, Settings> cronjobs : settings.getGroups("event.clock").entrySet()) {
            Settings entrySettings = cronjobs.getValue();
            if (entrySettings.getAsBoolean("enabled", true)) {
                String entry = entrySettings.get("entry");
                if (entry != null) {
                    try {
                        String className = entrySettings.get("class");
                        if (className != null) {
                            @SuppressWarnings("unchecked")
                            Class<? extends ClockEvent> eventClass = (Class<? extends ClockEvent>) classLoader.loadClass(className);
                            cronSchedule.add(className, CronExpression.parse(entry), new ClockEventService(eventBus, eventClass));
                            logger.log(Level.INFO, "cron job " + cronjobs.getKey() + " scheduled on " + entry + ", event class " + className);
                        } else {
                            logger.log(Level.WARNING, "no class specified");
                        }
                    } catch (Exception e) {
                        logger.log(Level.WARNING, "unable to schedule cron job " + cronjobs.getKey() + ", reason " + e.getMessage());
                    }
                }
            }
        }
        logger.log(Level.INFO, "entries = " + cronSchedule.getEntries());
        cronSchedule.start();
    }

    @Override
    public void close() throws IOException {
        cronSchedule.close();
    }
}
