package org.xbib.event.io.file;

import org.xbib.event.bus.AsyncEventBus;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.Objects;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.regex.Pattern;

public class FileFollowEventManager implements Closeable {

    private static final Logger logger = Logger.getLogger(FileFollowEventManager.class.getName());

    private final Map<Future<?>, FileFollowEventService> eventServiceMap;

    @SuppressWarnings("unchecked")
    public FileFollowEventManager(Settings settings,
                                  AsyncEventBus eventBus,
                                  ExecutorService executorService,
                                  ClassLoader classLoader) {
        this.eventServiceMap = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : settings.getGroups("event.filefollow").entrySet()) {
            Settings definition = entry.getValue();
            if (definition.getAsBoolean("enabled", true)) {
                String baseStr = definition.get("base");
                Objects.requireNonNull(baseStr);
                String patternStr = definition.get("pattern");
                Objects.requireNonNull(patternStr);
                try {
                    Path base = Paths.get(baseStr);
                    Pattern pattern = Pattern.compile(patternStr);
                    String className = definition.get("class", DefaultFileFollowEvent.class.getName());
                    Class<? extends FileFollowEvent> eventClass = (Class<? extends FileFollowEvent>) classLoader.loadClass(className);
                    FileFollowEventService fileFollowEventService = new FileFollowEventService(definition, eventBus, base, pattern, eventClass);
                    Future<?> future = executorService.submit(fileFollowEventService);
                    eventServiceMap.put(future, fileFollowEventService);
                    logger.log(Level.INFO, "file follow service " + entry.getKey() + " with base " + base + " and pattern " + pattern + " added, event class " + className);
                } catch (Exception e) {
                    logger.log(Level.SEVERE, "unable to create file follow service " + entry.getKey() + ", reason " + e.getMessage(), e);
                }
            }
        }
    }

    @Override
    public void close() throws IOException {
        for (Map.Entry<Future<?>, FileFollowEventService> entry : eventServiceMap.entrySet()) {
            entry.getValue().setKeepWatching(false);
            entry.getKey().cancel(true);
        }
    }
}
