package org.xbib.event.queue.path.simple;

import org.xbib.datastructures.json.tiny.Json;

import java.io.IOException;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.LinkedHashMap;
import java.util.Objects;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.event.bus.EventBus;

import static org.xbib.event.queue.path.simple.PathQueueEvent.INCOMING;

public class PathQueueEventService  {

    private static final Logger logger = Logger.getLogger(PathQueueEventService.class.getName());

    private final EventBus eventBus;

    private final Path path;

    private final int maxFileSize;

    private final Class<? extends PathQueueEvent> pathEventClass;

    private int eventCount;

    public PathQueueEventService(EventBus eventBus,
                                 Path path,
                                 int maxFileSize,
                                 Class<? extends PathQueueEvent> pathEventClass) throws IOException {
        this.eventBus = eventBus;
        this.path = path;
        this.maxFileSize = maxFileSize;
        this.pathEventClass = pathEventClass;
        drainIncoming();
        logger.log(Level.INFO, "at startup: watch service created for incoming files at " + path +
                "max file size = " + maxFileSize);
    }

    private void drainIncoming() throws IOException {
        try (DirectoryStream<Path> directoryStream = Files.newDirectoryStream(path.resolve(INCOMING))) {
            directoryStream.forEach(path -> {
                if (Files.isRegularFile(path)) {
                    String key = path.getFileName().toString();
                    logger.log(Level.INFO, "while draining found key=" + key + " path=" + path);
                    try {
                        postEvent(key, path);
                    } catch (IOException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                    }
                }
            });
        }
    }

    public void postEvent(String key, Path file) throws IOException {
        String base = getBase(key);
        String suffix = getSuffix(key);
        PathQueueEvent event;
        try {
            event = pathEventClass.getConstructor().newInstance();
        } catch (InstantiationException | IllegalAccessException | InvocationTargetException | NoSuchMethodException e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
            return;
        }
        event.setKey(base);
        event.setFile(file);
        event.setSuffix(suffix);
        event.setMap(new LinkedHashMap<>());
        event.setPath(path); // remember directory for fail() and success()
        long fileSize = Files.size(file);
        if (fileSize < maxFileSize && Objects.equals(suffix, "json")) {
            // we read only small JSON files
            try (Reader reader = Files.newBufferedReader(file)) {
                event.setMap(Json.toMap(reader));
            } catch (Exception e) {
                logger.log(Level.WARNING, e.getMessage(), e);
            }
            logger.log(Level.INFO, "posting new event =" + event.getClass() + " " + event);
            eventBus.post(event);
            eventCount++;
        } else {
            logger.log(Level.SEVERE, "skipping post event because incoming file is too large, max file size = " + maxFileSize);
        }
    }

    public static String getBase(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(0, pos) : name;
    }

    public static String getSuffix(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(pos + 1) : null;
    }
}
