package org.xbib.event.queue.path.watch;

import org.xbib.datastructures.api.TimeValue;
import org.xbib.datastructures.json.tiny.Json;
import org.xbib.event.bus.EventBus;
import org.xbib.settings.Settings;

import java.io.Closeable;
import java.io.IOException;
import java.io.Writer;
import java.nio.file.FileVisitResult;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.nio.file.SimpleFileVisitor;
import java.nio.file.attribute.BasicFileAttributes;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Future;
import java.util.logging.Level;
import java.util.logging.Logger;
import java.util.stream.Stream;

public class PathQueueEventManager implements Closeable {

    private static final Logger logger = Logger.getLogger(PathQueueEventManager.class.getName());

    private final Settings settings;

    private final Path path;

    private final List<Future<?>> futures;

    private final Map<Path, PathQueueEventService> pathEventServiceMap;

    @SuppressWarnings("unchecked")
    public PathQueueEventManager(Settings settings,
                                 EventBus eventBus,
                                 ExecutorService executorService,
                                 ClassLoader classLoader) {
        this.settings = settings;
        this.futures = new ArrayList<>();
        this.pathEventServiceMap = new LinkedHashMap<>();
        this.path = Paths.get(settings.get("path", "/var/tmp/default/queues"));
        for (Map.Entry<String, Settings> entry : settings.getGroups("queues").entrySet()) {
            String name = entry.getKey();
            Settings definition = entry.getValue();
            String eventClassName = definition.get("class");
            if (eventClassName == null) {
                throw new IllegalArgumentException("no path event class name given for path event manager at " + path);
            }
            try {
                Class<? extends PathQueueEvent> eventClass = (Class<? extends PathQueueEvent>) classLoader.loadClass(eventClassName);
                Path p = path.resolve(name);
                if (!Files.exists(p)) {
                    logger.log(Level.INFO, "creating queue " + name + " at " + p);
                    Files.createDirectories(p);
                }
                for (String string : List.of(PathQueueEvent.INCOMING, PathQueueEvent.SUCCESS, PathQueueEvent.FAIL)) {
                    Path dir = p.resolve(string);
                    if (!Files.exists(dir)) {
                        logger.log(Level.INFO, "creating queue " + name + " dir = " + dir);
                        Files.createDirectories(dir);
                    }
                }
                if (!pathEventServiceMap.containsKey(p)) {
                    int maxFileSize = settings.getAsInt("maxfilesize", 10 * 1024 * 1024); // max. 10 MB
                    logger.log(Level.INFO, "creating path event service on " + p + " with class " + eventClassName + " max file size = " + maxFileSize);
                    PathQueueEventService pathEventService = new PathQueueEventService(eventBus, p, maxFileSize, eventClass);
                    pathEventServiceMap.put(p, pathEventService);
                    futures.add(executorService.submit(pathEventService));
                }
            } catch (Exception e) {
                logger.log(Level.WARNING, "unable to create path event service " + eventClassName + " reason " + e.getMessage());
            }
        }
        logger.log(Level.INFO, pathEventServiceMap.size() + " path event services initialized");
    }

    public boolean put(String queue, String key, Map<String,Object> map) throws IOException {
        String keyFileName = key + ".json";
        Path queuePath = path.resolve(queue);
        if (Files.exists(queuePath.resolve(PathQueueEvent.INCOMING).resolve(keyFileName)) ||
                Files.exists(queuePath.resolve(PathQueueEvent.SUCCESS).resolve(keyFileName))) {
            logger.log(Level.WARNING, "key " + key + " already exists");
            return false;
        }
        try (Writer writer = Files.newBufferedWriter(queuePath.resolve(PathQueueEvent.INCOMING).resolve(keyFileName))) {
            writer.write(Json.toString(map));
        }
        purge(queue);
        return true;
    }

    public boolean put(String queue, String key, String value) throws IOException {
        String keyFileName = key + ".txt";
        Path queuePath = path.resolve(queue);
        if (Files.exists(queuePath.resolve(PathQueueEvent.INCOMING).resolve(keyFileName)) ||
                Files.exists(queuePath.resolve(PathQueueEvent.SUCCESS).resolve(keyFileName))) {
            logger.log(Level.WARNING, "key " + key + " already exists");
            return false;
        }
        try (Writer writer = Files.newBufferedWriter(queuePath.resolve(PathQueueEvent.INCOMING).resolve(keyFileName))) {
            writer.write(value);
        }
        purge(queue);
        return true;
    }

    public boolean exists(String queue, String key) {
        String keyFileName = key + ".json";
        Path queuePath = path.resolve(queue);
        return Files.exists(queuePath.resolve(PathQueueEvent.INCOMING).resolve(keyFileName)) ||
                Files.exists(queuePath.resolve(PathQueueEvent.SUCCESS).resolve(keyFileName));
    }

    public long sizeOfIncoming(String queue) throws IOException {
        Path queuePath = path.resolve(queue);
        Path incomingPath = queuePath.resolve(PathQueueEvent.INCOMING);
        return sizeOf(incomingPath);
    }

    public long sizeOfSuccess(String queue) throws IOException {
        Path queuePath = path.resolve(queue);
        Path successPath = queuePath.resolve(PathQueueEvent.SUCCESS);
        return sizeOf(successPath);
    }

    public static long sizeOf(Path path) throws IOException {
        Stream<Path> stream = Files.find(path, 1, (p, basicFileAttributes) -> Files.isRegularFile(p));
        return stream.count();
    }

    public void purge(String queue) throws IOException {
        List<String> purgeQueues = Arrays.asList(settings.getAsArray("purge"));
        if (purgeQueues.contains(queue)) {
            TimeValue lifetime = settings.getAsTime("lifetime", TimeValue.timeValueHours(72));
            Path queuePath = path.resolve(queue);
            Files.walkFileTree(queuePath, new SimpleFileVisitor<Path>() {
                @Override
                public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                    if (Instant.now().minusMillis(attrs.lastModifiedTime().toMillis()).toEpochMilli() > lifetime.millis()) {
                        logger.log(Level.WARNING, "lifetime " + lifetime + " exceeded, deleting  " + file);
                        Files.delete(file);
                    }
                    return FileVisitResult.CONTINUE;
                }
            });
        }
    }

    @Override
    public void close() throws IOException {
        logger.log(Level.INFO, "closing all path event services");
        for (PathQueueEventService service : pathEventServiceMap.values()) {
            logger.log(Level.INFO, "closing path event service " + service);
            service.close();
        }
        for (Future<?> future : futures) {
            future.cancel(true);
        }
    }
}
