package org.xbib.event.queue.path.watch;

import org.xbib.datastructures.json.tiny.Json;

import java.io.Closeable;
import java.io.IOException;
import java.io.Reader;
import java.lang.reflect.InvocationTargetException;
import java.nio.file.ClosedWatchServiceException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.StandardWatchEventKinds;
import java.nio.file.WatchEvent;
import java.nio.file.WatchKey;
import java.nio.file.WatchService;
import java.util.LinkedHashMap;
import java.util.Objects;
import java.util.concurrent.Callable;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.xbib.event.bus.EventBus;

import static org.xbib.event.queue.path.watch.PathQueueEvent.INCOMING;

public class PathQueueEventService implements Callable<Integer>, Closeable {

    private static final Logger logger = Logger.getLogger(PathQueueEventService.class.getName());

    private final EventBus eventBus;

    private final Path path;

    private final int maxFileSize;

    private final Class<? extends PathQueueEvent> pathEventClass;

    private final WatchService watchService;

    private int eventCount;

    private volatile boolean keepWatching;

    public PathQueueEventService(EventBus eventBus,
                                 Path path,
                                 int maxFileSize,
                                 Class<? extends PathQueueEvent> pathEventClass) throws IOException {
        this.eventBus = eventBus;
        this.path = path;
        this.maxFileSize = maxFileSize;
        this.pathEventClass = pathEventClass;
        drainIncoming();
        this.watchService = path.getFileSystem().newWatchService();
        WatchEvent.Kind<?>[] kinds = new WatchEvent.Kind<?>[] { StandardWatchEventKinds.ENTRY_CREATE };
        WatchKey watchKey = path.resolve(INCOMING).register(watchService, kinds);
        keepWatching = true;
        logger.log(Level.INFO, "at startup: watch service created for incoming files at " + path +
                "max file size = " + maxFileSize);
    }

    @Override
    public Integer call() {
        try {
            logger.log(Level.INFO, "watch service running on " + path.resolve(INCOMING));
            while (keepWatching && watchService != null) {
                WatchKey watchKey = watchService.take();
                logger.log(Level.FINE, "received a watch key " + watchKey);
                for (WatchEvent<?> watchEvent : watchKey.pollEvents()) {
                    WatchEvent.Kind<?> kind = watchEvent.kind();
                    if (kind == StandardWatchEventKinds.OVERFLOW) {
                        continue;
                    }
                    // we sleep here a bit, to give time to the OS to complete file writing
                    Thread.sleep(1000L);
                    @SuppressWarnings("unchecked")
                    WatchEvent<Path> pathWatchEvent = (WatchEvent<Path>) watchEvent;
                    String watchEventContext = pathWatchEvent.context().toString();
                    Path p = path.resolve(INCOMING).resolve(watchEventContext);
                    logger.log(Level.FINE, "watch event" + pathWatchEvent + " key=" + watchEventContext + " path=" + p);
                    postEvent(watchEventContext, p);
                }
                watchKey.reset();
            }
        } catch (ClosedWatchServiceException e) {
            logger.log(Level.FINE, "closed watch key");
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
            logger.log(Level.FINE, "interrupted: " + e.getMessage());
        } catch (Exception e) {
            logger.log(Level.SEVERE, e.getMessage(), e);
        }
        return eventCount;
    }

    @Override
    public void close() throws IOException {
        keepWatching = false;
        if (watchService != null) {
            logger.log(Level.FINE, "closing watch service " + watchService);
            watchService.close();
        }
    }

    private void drainIncoming() throws IOException {
        try (DirectoryStream<Path> directoryStream = Files.newDirectoryStream(path.resolve(INCOMING))) {
            directoryStream.forEach(path -> {
                if (Files.isRegularFile(path)) {
                    String key = path.getFileName().toString();
                    logger.log(Level.INFO, "while draining found key=" + key + " path=" + path);
                    try {
                        postEvent(key, path);
                    } catch (IOException | NoSuchMethodException | InvocationTargetException | InstantiationException | IllegalAccessException e) {
                        logger.log(Level.SEVERE, e.getMessage(), e);
                    }
                }
            });
        }
    }

    private void postEvent(String key, Path file)
            throws NoSuchMethodException, InvocationTargetException, InstantiationException, IllegalAccessException, IOException {
        String base = getBase(key);
        String suffix = getSuffix(key);
        PathQueueEvent event = pathEventClass.getConstructor().newInstance();
        event.setKey(base);
        event.setFile(file);
        event.setSuffix(suffix);
        event.setMap(new LinkedHashMap<>());
        event.setPath(path); // remember directory for fail() and success()
        long fileSize = Files.size(file);
        if (fileSize < maxFileSize && Objects.equals(suffix, "json")) {
            // we read only small JSON files
            try (Reader reader = Files.newBufferedReader(file)) {
                event.setMap(Json.toMap(reader));
            } catch (Exception e) {
                logger.log(Level.WARNING, e.getMessage(), e);
            }
            logger.log(Level.INFO, "posting new event =" + event.getClass() + " " + event);
            eventBus.post(event);
            eventCount++;
        } else {
            logger.log(Level.SEVERE, "skipping post event because incoming file is too large, max file size = " + maxFileSize);
        }
    }

    public static String getBase(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(0, pos) : name;
    }

    public static String getSuffix(String name) {
        int pos = name.lastIndexOf('.');
        return pos >= 0 ? name.substring(pos + 1) : null;
    }
}
