package org.xbib.event.timer;

import org.xbib.event.bus.AsyncEventBus;
import org.xbib.event.bus.EventBus;
import org.xbib.event.persistence.FilePersistenceStore;
import org.xbib.event.persistence.PersistenceStore;
import org.xbib.settings.Settings;
import org.xbib.time.chronic.Chronic;
import org.xbib.time.chronic.Span;

import java.io.Closeable;
import java.io.IOException;
import java.text.ParseException;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.util.LinkedHashMap;
import java.util.Map;
import java.util.concurrent.Executors;
import java.util.logging.Level;
import java.util.logging.Logger;

public class TimerEventManager implements Closeable {

    private static final Logger logger = Logger.getLogger(TimerEventManager.class.getName());

    private final Map<String, TimerEventService> services;

    public TimerEventManager(Settings settings) {
        this(settings, new AsyncEventBus(Executors.newSingleThreadExecutor()),
                TimerEventManager.class.getClassLoader(), ZoneId.systemDefault());
    }

    @SuppressWarnings("unchecked")
    public TimerEventManager(Settings settings,
                             EventBus eventBus,
                             ClassLoader classLoader,
                             ZoneId zoneId) {
        this.services = new LinkedHashMap<>();
        for (Map.Entry<String, Settings> entry : settings.getGroups("event.timer").entrySet()) {
            String name = entry.getKey();
            Settings timerSettings = entry.getValue();
            String className = timerSettings.get("class", TimerEvent.class.getName());
            try {
                Class<? extends TimerEvent> eventClass = (Class<? extends TimerEvent>) classLoader.loadClass(className);
                PersistenceStore<String, Object> persistenceStore = new FilePersistenceStore(timerSettings, name);
                services.put(name, new TimerEventService(eventBus, name, eventClass, zoneId, persistenceStore));
                logger.log(Level.INFO, "timer " + name + " active for timer event class " + className);
            } catch (Exception e) {
                logger.log(Level.WARNING, "unable to activate timer " + name + ", reason: " + e.getMessage(), e);
            }
        }
    }

    public boolean put(String key, String timeSpec, Map<String,Object> map) throws ParseException, IOException {
        if (services.containsKey(key)) {
            Span span = Chronic.parse(timeSpec);
            if (span != null) {
                ZonedDateTime zonedDateTime = span.getBeginCalendar();
                services.get(key).schedule(zonedDateTime.toInstant(), map);
                logger.log(Level.INFO, "scheduled to " + zonedDateTime);
            } else {
                logger.log(Level.INFO, "timer event key " + key + ": can not understand time spec " + timeSpec);
                return false;
            }
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + key);
        }
        return false;
    }

    public boolean put(String key, Instant instant, Map<String,Object> map) throws IOException {
        if (services.containsKey(key)) {
            services.get(key).schedule(instant, map);
            return true;
        } else {
            logger.log(Level.SEVERE, "unknown timer event key: " + key);
        }
        return false;
    }

    public void purge() {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "purging timer " + entry.getKey());
            entry.getValue().purge();
        }
    }

    @Override
    public void close() throws IOException {
        for (Map.Entry<String, TimerEventService> entry : services.entrySet()) {
            logger.log(Level.INFO, "closing timer " + entry.getKey());
            entry.getValue().close();
        }
    }
}
